@@ -34,32 +34,42 @@ ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 """
 
 import json
-import mdbf as dbf
 import os
 import re
+import tempfile
+import subprocess 
+import csv
 
 from collections import defaultdict
-from datetime import date
+from datetime import date, datetime
 
 class FieldError(Exception):
     pass
 
+class HarbourError(Exception):
+    pass
+
+class FileError(Exception):
+    pass
+
 class Type:
     NULL = ord('?') # no type inferrence has been attempted
     UNDEFINED = ord('X') # type inferrence failed
-    CHAR = int(dbf.FieldType.CHAR)
-    CURRENCY = int(dbf.FieldType.CURRENCY)
-    DATE = int(dbf.FieldType.DATE)
-    DATETIME = int(dbf.FieldType.DATETIME)
-    DOUBLE = int(dbf.FieldType.DOUBLE)
-    FLOAT = int(dbf.FieldType.FLOAT)
-    GENERAL = int(dbf.FieldType.GENERAL)
-    INTEGER = int(dbf.FieldType.INTEGER)
-    LOGICAL = int(dbf.FieldType.LOGICAL)
-    MEMO = int(dbf.FieldType.MEMO)
-    NUMERIC = int(dbf.FieldType.NUMERIC)
-    PICTURE = int(dbf.FieldType.PICTURE)
-    TIMESTAMP = int(dbf.FieldType.TIMESTAMP)
+    CHAR = ord('C')
+    CURRENCY = ord('Y')
+    DATE = ord('D')
+    DATETIME = ord('T')
+    DOUBLE = ord('B')
+    FLOAT = ord('F')
+    GENERAL = ord('G')
+    INTEGER = ord('I')
+    LOGICAL = ord('L')
+    MEMO = ord('M')
+    NUMERIC = ord('N')
+    PICTURE = ord('P')
+    TIMESTAMP = ord('@')
+
+_FNAME, _COMP, _VALUE = 0, 1, 2
 
 class Field:
     """Field class represents the meta DBF header field"""
@@ -78,6 +88,20 @@ class Field:
     def is_type(self, ftype):
         return self.ftype == ftype
 
+    @staticmethod
+    def dtoiso(value):
+        if isinstance(value, date) or isinstance(value, datetime):
+            return value.isoformat()
+        else:
+            return date(int(value[:4]), int(value[4:6]), int(value[6:8])).isoformat()
+
+    @staticmethod
+    def isotod(value):
+        if isinstance(value, date) or isinstance(value, datetime):
+            return value.strftime('%Y%m%d')
+        else:
+            return date.fromisoformat(value).strftime('%Y%m%d')
+
     def _pad(self, value, length=None, pad=None, side=None):
         length = self.length if not length else length
         pad = self.pad if not pad else pad
@@ -102,6 +126,10 @@ class Field:
         """converts ftype values into ctype if possible"""
         if self.ctype == Type.NULL or self.ctype == Type.UNDEFINED:
             raise ValueError('missing inferred type for Field %s' % self.name)
+        if self.ctype == Type.LOGICAL:
+            if isinstance(value, str):
+                return True if value == 'T' else False
+            return value
         if self.ctype == Type.INTEGER:
             return int(value) if value else None
         if self.ctype == Type.CHAR:
@@ -109,40 +137,229 @@ class Field:
         if self.ctype == Type.FLOAT:
             return value
         if self.ctype == Type.DATE:
-            return value.isoformat() if value else value
+            return Field.dtoiso(value) if value and isinstance(value, str) and len(value) == 8 else value
         if self.ftype == self.ctype:
             return value
         raise ValueError('undefined conversion method from ftype %s to ctype %s' % (chr(self.ftype), chr(self.ctype)))
 
     def ctof(self, value):
         """converst ctype values back into native ftype with padding and all"""
+        if value != None:
+            if self.ftype == Type.INTEGER:
+                return int(value)
+            if self.ftype == Type.NUMERIC:
+                return float(value)
+            if self.ftype == Type.CHAR:
+                if self.is_padded:
+                    return self._pad(str(value))
+                else:
+                    return str(value)
+            if self.ftype == Type.DATE:
+                if isinstance(value, str):
+                    if len(value) == 8:
+                        return Field.dtoiso(value)
+                    if len(value) == 10:
+                        return value
+                raise ValueError('Incorrect date value: %s' % value)
+            if self.ftype == Type.LOGICAL:
+                if isinstance(value, str):
+                    return True if value == 'T' else False
+        return value
+
+    @staticmethod
+    def quote(value, char='"'):
+        return char + value.replace(char, f'\\{char}') + char
+
+    def ctox(self, value):
+        """converst ctype values into export csv compatible values"""
+        if self.ftype == Type.LOGICAL:
+            return 'T' if value else 'F'
         if self.ftype == Type.CHAR:
             if self.ctype == Type.INTEGER:
                 if isinstance(value, int):
-                    if self.is_padded:
-                        return self._pad(str(value))
+                    if self.is_padded and self.is_padded != 'R':
+                        return Field.quote(self._pad(str(value)))
+                if value != None:
+                    return Field.quote(str(value))
+            if self.ctype == Type.CHAR:
+                if isinstance(value, str):
+                    if self.is_padded and self.is_padded != 'R':
+                        return Field.quote(self._pad(value))
                     else:
-                        return str(value)
-            elif self.ctype == Type.CHAR:
-                if self.is_padded and isinstance(value, str):
-                    return self._pad(value)
+                        return Field.quote(value)
         if self.ftype == Type.DATE:
             if isinstance(value, str):
-                return date.fromisoformat(value)
-        return value
+                return Field.isotod(value)
+            if isinstance(value, date) or isinstance(value, datetime):
+                return Field.isotod(value)
+        try:
+            return str(value) if value != None else Field.quote('')
+        except:
+            return Field.quote('')
+
+class MetaRecord:
+    def __init__(self, header_fields, record):
+        self._header_fields = header_fields
+        self._record = record
+
+    def __getitem__(self, key):
+        if isinstance(key, str):
+            if key in self._header_fields:
+                return self._record[self._header_fields.index(key)]
+            else:
+                raise IndexError
+        elif isinstance(key, int):
+            if key < len(self._header_fields):
+                return self._record[key]
+            else:
+                raise IndexError
+        raise TypeError
+
+class DBFIterator:
+    def __init__(self, records, header_fields):
+        self._records = records
+        self._header_fields = header_fields
+
+    def __iter__(self):
+        return self
+
+    def __next__(self):
+        if len(self._records) > 0:
+            return MetaRecord(self._header_fields, self._records.pop(0))
+        else:
+            raise StopIteration
+
+_EXE = 'dbfadapter'
+
+class DBFAdapter:
+    def __init__(self, db_path, table_name, mode='-', index_suffix='ntx'):
+        if os.path.isfile(db_path + table_name):
+            self._table = self
+            self.db_path = db_path
+            self.table_name = table_name
+            self.index_suffix = index_suffix
+            self.index_files = self._locate_index_files()
+            self._records = None
+            self._meta = None
+            self._filter = []
+        else:
+            raise FileError('No such database exists.')
+
+    @staticmethod
+    def _head(db_path, table_name, _EXE=_EXE):
+        table_name = table_name.split('.')[0]
+        headers, f, fd, fname = None, None, None, None
+        try:
+            fd, fname = tempfile.mkstemp(prefix='dbfh', suffix='.json', text=True)
+            ext = subprocess.run([_EXE, "head", db_path, table_name, fname, '//noalert'], timeout=10, text=True, capture_output=True)
+            if ext.returncode != 0:
+                raise HarbourError(ext.stderr)
+            f = os.fdopen(fd, 'r')
+            headers = f.read()
+        finally:
+            if f:
+                f.close()
+            if fname:
+                os.remove(fname)
+        return headers
+
+    @staticmethod
+    def _export(db_path, table_name, index_files, where=[], _EXE=_EXE):
+        table_name = table_name.split('.')[0]
+        records, f, fd, fname = [], None, None, None
+        try:
+            fd, fname = tempfile.mkstemp(prefix='dbfx', suffix='.cson', text=True)
+            with open(fname, 'w') as f:
+                json.dump(where, f)
+            ext = subprocess.run([_EXE, "export", db_path, table_name, fname, *index_files, '//noalert'], timeout=10, text=True, capture_output=True)
+            if ext.returncode != 0:
+                raise HarbourError(ext.stderr)
+            with open(fname, 'r') as f:
+                csvf = csv.reader(f)
+                records = [r for r in csvf]
+        finally:
+            if f:
+                f.close()
+            if fname:
+                os.remove(fname)
+        return records[:-1]
+
+    @staticmethod
+    def _append(line, db_path, table_name, index_files, _EXE=_EXE):
+        table_name = table_name.split('.')[0]
+        f, fd, fname = None, None, None
+        try:
+            fd, fname = tempfile.mkstemp(prefix='dbfa', suffix='.csv', text=True)
+            f = os.fdopen(fd, 'w')
+            f.write(','.join(line))
+            f.close()
+            ext = subprocess.run([_EXE, "append", db_path, table_name, fname, *index_files, '//noalert'], timeout=10, text=True, capture_output=True)
+            if ext.returncode != 0:
+                raise HarbourError(ext.stderr)
+        finally:
+            if f:
+                f.close()
+            if fname:
+                os.remove(fname)
+        return True
+
+    @staticmethod
+    def _update(update_package, db_path, table_name, index_files, _EXE=_EXE):
+        table_name = table_name.split('.')[0]
+        f, fd, fname, updated_count = None, None, None, None
+        try:
+            fd, fname = tempfile.mkstemp(prefix='dbfu', suffix='.json', text=True)
+            f = os.fdopen(fd, 'w')
+            data = json.dump(update_package, f)
+            f.close()
+            ext = subprocess.run([_EXE, "update", db_path, table_name, fname, *index_files, '//noalert'], timeout=10, text=True, capture_output=True)
+            updated_count = ext.stdout.split('\n')
+            updated_count = updated_count[0].split(':') if len(updated_count) > 0 else None
+            updated_count = updated_count[1] if len(updated_count) > 0 else None
+            if ext.returncode != 0:
+                raise HarbourError(ext.stderr)
+        finally:
+            if f:
+                f.close()
+            if fname:
+                os.remove(fname)
+        return int(updated_count)
+
+    def _locate_index_files(self):
+        """Finds index files for db if they exist, **VERY MUCH** case sensitive"""
+        indices = []
+        with os.scandir(self.db_path) as d:
+            for e in d:
+                if e.is_file():
+                    if e.name.startswith(self.table_name.split('.')[0]):
+                        if e.name.lower().endswith('.' + self.index_suffix.lower()):
+                            indices.append(e.name.split('.')[0])
+        return indices
+
+    def _parse_meta(self):
+        header = json.loads(DBFAdapter._head(self.db_path, self.table_name))
+        header = [x[1:-1].split(',') for x in header]
+        header = [[y.strip() for y in x] for x in header]
+        self._meta = { x[0]: (x[0], ord(x[1]), int(x[2]), int(x[3])) for x in header }
+
+    def __iter__(self):
+        if not self._meta:
+            self._parse_meta()
+        records = DBFAdapter._export(self.db_path, self.table_name, self.index_files, self._filter)
+        return DBFIterator(records, [k for k, v in self._meta.items() if v[1] != Type.MEMO] )
+
+    def field_info(self, field_name):
+        return self._meta[field_name][1:] if self._meta else None
 
 # @TODO-EP-002: Determine mandatory header fields
-class RFKAdapter:
-    def __init__(self, db_path, table_name, mode='-'):
-        self.db_path = db_path
-        self.table_name = table_name
-        self._table = dbf.Table(db_path + table_name, codepage='cp852') # dbf_type='db3'
-        self._table.open(mode=dbf.READ_WRITE if mode.lower() == 'w' else dbf.READ_ONLY)
-        self._parse_headers()
-
-    def __del__(self):
-        if hasattr(self, '_table') and self._table.status != dbf.CLOSED:
-            self._table.close()
+class RFKAdapter(DBFAdapter):
+    def __init__(self, db_path, table_name, mode='-', index_suffix='ntx', with_headers=True):
+        super(RFKAdapter, self).__init__(db_path, table_name, mode, index_suffix)
+        self._table = self
+        base_name = os.path.splitext(self.table_name)[0]
+        self._cache_path = os.path.join(self.db_path, base_name + '.json')
+        if with_headers:
+            self._parse_headers()
 
     @staticmethod
     def _prepare_value(value):
@@ -255,12 +472,13 @@ class RFKAdapter:
                 return False, None
         return is_strings, None
 
-    def _parse_headers(self):
+    def _parse_headers(self, flush=False):
         """Parses the fields from table headers"""
-        self.header_fields = { _name: Field(_name, _prop[0], _prop[2], _prop[4]) for _name, _prop in self._table._meta.items() }
-        base_name = os.path.splitext(self.table_name)[0]
-        cache_path = os.path.join(self.db_path, base_name + '.json')
-        if os.path.isfile(cache_path):
+        if flush and os.path.isfile(self._cache_path):
+            self._flush_headers()
+        self._parse_meta()
+        self.header_fields = { _name: Field(_name, _prop[1], _prop[2], _prop[3]) for _name, _prop in self._table._meta.items() if _prop[1] != Type.MEMO }
+        if os.path.isfile(self._cache_path):
             self._restore_headers()
         else:
             for field in self.header_fields.values():
@@ -279,6 +497,7 @@ class RFKAdapter:
                     field.ctype = Type.FLOAT
                 else:
                     field.ctype = field.ftype
+            self._cache_headers()
 
     def _read(self, where=[], raw_result=False, infer_type=False):
         """Read, fetch and filter from RFK table
@@ -288,38 +507,32 @@ class RFKAdapter:
         - `infer_type` flag forces using inferred types everywhere
         """
         result = []
-        fields = self._table._meta.fields
-        for field_name, constr in where:
-            if field_name not in fields:
-                raise FieldError('No field with name %s in table %s' % (field_name, self.table_name))
-                return
+        fields = [k for k, v in self._meta.items() if v[1] != Type.MEMO]
+
+        self._filter = self._convert_conditions(where)
+
         for record in self._table:
-            satisfies = True
-            for field_name, constr in where:
-                value = RFKAdapter._prepare_value(record[field_name])
-                if infer_type:
-                    value = self.header_fields[field_name].ftoc(value)
-                if not constr(value):
-                    satisfies = False
-                    break
-            if satisfies:
-                result_record = {}
-                for field_name in fields:
-                    result_value = record[field_name]
-                    if not raw_result:
-                        result_value = RFKAdapter._prepare_value(result_value)
-                        if infer_type:
-                            try:
-                                result_value = self.header_fields[field_name].ftoc(result_value)
-                            except ValueError:
-                                pass
-                    result_record[field_name] = result_value
-                result.append(result_record)
+            result_record = {}
+            for field_name in fields:
+                result_value = record[field_name]
+                if not raw_result:
+                    result_value = RFKAdapter._prepare_value(result_value)
+                    if infer_type:
+                        try:
+                            result_value = self.header_fields[field_name].ftoc(result_value)
+                        except ValueError:
+                            pass
+                result_record[field_name] = result_value
+            result.append(result_record)
         return result
 
+    def read(self, where=[]):
+        """Opinionated type inferred read"""
+        return self._read(where=where, infer_type=True)
+
     def read_all(self):
         """Returns all the DBF values, type inferred"""
-        return self._read()
+        return self._read(infer_type=True)
 
     def filter(self, where=[]):
         """Returns all filtered DBF values, best effort type inferred
@@ -328,91 +541,83 @@ class RFKAdapter:
         """
         return self._read(where, infer_type=True)
 
-    def _convert_condition(self, column, comparator, constraint):
-        """Converts between convenience and filter condition styles
+    def _convert_conditions(self, conditions):
+        """Converts between convenience and JSON filter condition styles
 
         < (lt), > (gt), <= (lte), >= (gte), == (eq), != (neq)
         si - parcijalno uparivanje stringova bez obzira na mala i velika slova
         s - parcijalno uparivanje stringova uz razlikovanje malih i velikih slova
         x - uparivanje stringova korištenjem regularnih izraza (regex)
         """
-        _map = {
-            'lt': lambda x: x < self.header_fields[column].strtoc(constraint),
-            'gt': lambda x: x > self.header_fields[column].strtoc(constraint),
-            'lte': lambda x: x <= self.header_fields[column].strtoc(constraint),
-            'gte': lambda x: x >= self.header_fields[column].strtoc(constraint),
-            'eq': lambda x: x == self.header_fields[column].strtoc(constraint),
-            'neq': lambda x: x != self.header_fields[column].strtoc(constraint),
-            'si': lambda x: x.lower().find(constraint.lower()) >= 0,
-            's': lambda x: x.find(constraint) >= 0,
-            'x': lambda x: re.search(constraint, x) != None,
-        }
-        if comparator in ['si', 's', 'x'] and self.header_fields[column].ctype != Type.CHAR:
-            raise ValueError('invalid compararator for non string column type')
-        if comparator in _map:
-            return (column, _map[comparator])
-        else:
-            raise ValueError('invalid comparator %s' % comparator)
+        _where = []
+        _comparators = ['lt', 'gt', 'lte', 'gte', 'eq', 'neq', 'si', 's', 'x']
+        for condition in conditions:
+            if len(condition) != 3:
+                raise TypeError('Invalid filter. (NOTE: Old style lambda filters are deprecated, use convenience filters)')
+            if condition[_FNAME] not in self.header_fields.keys():
+                raise FieldError('No field with name %s in table %s' % (condition[_FNAME], self.table_name))
+            if condition[_COMP] in ['si', 's', 'x'] and self.header_fields[condition[_FNAME]].ctype != Type.CHAR:
+                raise ValueError('invalid comparator for non string column type')
+            if condition[_COMP] in _comparators:
+                _where.append({
+                    'column_name': condition[_FNAME],
+                    'comparator': condition[_COMP],
+                    'value': self.header_fields[condition[_FNAME]].ctof(condition[_VALUE])})
+            else:
+                raise ValueError('Invalid filter condition.')
+        return _where
 
-    def where(self, conditions=[]):
+    def where(self, where=[]):
         """Convenience filter method
 
         Written in order to facilitate external calls with conditions
         styled like: [('COL', 'gt', 'VALUE'), ('COL', 'lt', 'VALUE')]
-        conditions get type inferred and operators go lambdas
+        conditions get type inferred
         """
-        converted_conditions = [self._convert_condition(*c) for c in conditions]
-        return self.filter(converted_conditions)
+        return self.filter(where)
 
     def write(self, data):
         """Appends a new record to the table"""
-        for k, v in data.items():
-            data[k] = self.header_fields[k].ctof(v)
-        self._table.append(data)
+        line = []
+        for _, field in self.header_fields.items():
+            if field.name in data:
+                if field.ftype != Type.MEMO:
+                    line.append(field.ctox(data[field.name]))
+            else:
+                line.append(field.ctox(None))
+        self._append(line, self.db_path, self.table_name, self.index_files)
 
     def update(self, what, where):
         """Updates existing records, returns True on success"""
-        _where = []
-        for c in where:
-            if len(c) == 2:
-                _where.append(c)
-            if len(c) == 3:
-                _where.append(self._convert_condition(*c))
-        fields = self._table._meta.fields
-        for field_name, constr in _where:
-            if field_name not in fields:
-                raise FieldError('No field with name %s in table %s' % (field_name, self.table_name))
-        for record in self._table:
-            satisfies = True
-            for field_name, constr in _where:
-                value = RFKAdapter._prepare_value(record[field_name])
-                value = self.header_fields[field_name].ftoc(value)
-                if not constr(value):
-                    satisfies = False
-                    break
-            if satisfies:
-                for k, v in what.items():
-                    with record:
-                        record[k] = self.header_fields[k].ctof(v)
-        return True
+        _what = {}
+        _dict_where = self._convert_conditions(where)
+        if not self.header_fields:
+            raise ValueError('header_fields not initialized, run self._parse_headers first!')
+        for field_name, new_value in what.items():
+            if field_name not in self.header_fields.keys():
+                raise FieldError('No field with name %s in table %s to update' % (field_name, self.table_name))
+            _what[field_name] = self.header_fields[field_name].ctof(new_value)
+        return DBFAdapter._update({'what': _what, 'where': _dict_where}, self.db_path, self.table_name, self.index_files)
 
     def _cache_headers(self):
         """Caches parsed headers to file because parsing is time demanding
 
         @TODO-EP-001: update cache on structural changes, ATM this is unneccessary,
-        flush manually by deleting *.json
+        flush manually by deleting *.json or calling _parse_headers with flush_cache
         """
-        base_name = os.path.splitext(self.table_name)[0]
-        json_path = os.path.join(self.db_path, base_name + '.json')
-        with open(json_path, 'w') as fp:
+        with open(self._cache_path, 'w') as fp:
             headers = { field.name: field.__dict__ for field in self.header_fields.values() }
             json.dump(headers, fp)
 
+    def _flush_headers(self):
+        self.header_fields = None
+        if os.path.isfile(self._cache_path):
+            os.remove(self._cache_path)
+
     def _restore_headers(self):
         """Restores header fields from cache"""
-        base_name = os.path.splitext(self.table_name)[0]
-        json_path = os.path.join(self.db_path, base_name + '.json')
-        with open(json_path, 'r') as fp:
+        with open(self._cache_path, 'r') as fp:
             headers = json.load(fp)
+            self.header_fields = {}
             for field_name, field in headers.items():
                 self.header_fields[field_name] = Field(**field)
\ No newline at end of file
